# Written by Bram Cohen
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

from random import randrange, shuffle, randint
from time import clock


class Choker:
    def __init__(self, dow, config, schedule, picker, piecelength, readbufferallowedfunc, done = lambda: False):
        self.dow = dow
        self.config = config
        self.round_robin_period = config['round_robin_period']
        self.schedule = schedule
        self.picker = picker
        self.piecelength = piecelength
        self.readbufferallowedfunc = readbufferallowedfunc
        self.connections = []
        self.last_preferred = 0
        self.last_round_robin = clock()
        self.done = done
        self.super_seed = False
        self.paused = False
        if self.config['read_buffer_memory_saving'] == 0 and self.readbufferallowedfunc() < 0:
            self.nbuploads = self.oldmaxuploads = self.config['floor_uploads']
        else:
            self.nbuploads = self.oldmaxuploads = self.config['max_uploads']
        # To somewhat distribute torrents responses in time
        # and wait for piece buffers allocation to react
        self.limitinguploadscycle = self.lucinit = randint(0, 11)
        schedule(self._round_robin, 5)

    def updateData(self, piecelength):
        self.piecelength = piecelength

    def set_round_robin_period(self, x):
        self.round_robin_period = x

    def _round_robin(self):
        self.schedule(self._round_robin, 5)

        if len(self.connections) >= self.config['floor_uploads'] and self.config['read_buffer_memory_saving'] == 0:
            maxuploads = self.config['max_uploads']
            if maxuploads != self.oldmaxuploads:
                if self.nbuploads > maxuploads or self.nbuploads >= self.oldmaxuploads:
                    self.nbuploads = maxuploads
                self.limitinguploadscycle = 1
            elif self.limitinguploadscycle % 12 == 0:
                readbuffermargin = self.readbufferallowedfunc()
                if readbuffermargin >= self.piecelength and self.nbuploads < maxuploads:
                    self.nbuploads += 1
                elif readbuffermargin < 0 and self.nbuploads > self.config['floor_uploads']:
                    self.nbuploads -= 1
                self.limitinguploadscycle = 1
            else:
                self.limitinguploadscycle += 1
            minuploads = min(self.config['min_uploads'], self.nbuploads)
            self.oldmaxuploads = maxuploads
        else:
            self.nbuploads = self.config['max_uploads']
            minuploads = self.config['min_uploads']
            self.limitinguploadscycle = self.lucinit

        if self.super_seed:
            cons = range(len(self.connections))
            to_close = []
            count = minuploads - self.last_preferred
            if count > 0:   # optimization
                shuffle(cons)
            for c in cons:
                i = self.picker.next_have(self.connections[c], count > 0)
                if i is None:
                    continue
                if i < 0:
                    to_close.append(self.connections[c])
                    continue
                self.connections[c].send_have(i)
                count -= 1
            for c in to_close:
                c.close()
        if self.last_round_robin + self.round_robin_period < clock():
            self.last_round_robin = clock()
            for i in xrange(1, len(self.connections)):
                c = self.connections[i]
                u = c.upload
                if u.is_choked() and u.is_interested():
                    self.connections = self.connections[i:] + self.connections[:i]
                    break
        self._rechoke()

    def _rechoke(self):
        preferred = []
        if self.paused or not self.nbuploads:
            for c in self.connections:
                c.upload.choke()
            return
        if self.nbuploads > 1:
            for c in self.connections:
                u = c.upload
                if not u.is_interested():
                    continue
                if self.done():
                    r = u.get_rate()
                else:
                    d = c.download
                    r = d.get_rate()
                    if r < 1024 or d.is_snubbed():
                        continue
                preferred.append((-r, c))
            self.last_preferred = len(preferred)
            preferred.sort()
            del preferred[self.nbuploads - 1:]
            preferred = [x[1] for x in preferred]
        count = len(preferred)
        hit = False
        to_unchoke = []
        for c in self.connections:
            u = c.upload
            if c in preferred:
                to_unchoke.append(u)
            else:
                if count < self.nbuploads or not hit:
                    to_unchoke.append(u)
                    if u.is_interested():
                        count += 1
                        hit = True
                else:
                    u.choke()
        for u in to_unchoke:
            u.unchoke()

    def connection_made(self, connection, p = None):
        if p is None:
            p = randrange(-2, len(self.connections) + 1)
        self.connections.insert(max(p, 0), connection)
        self._rechoke()

    def connection_lost(self, connection, shutdown = False):
        self.connections.remove(connection)
        self.picker.lost_peer(connection)
        u = connection.upload
        if not shutdown and u.is_interested() and not u.is_choked():
            self._rechoke()

    def interested(self, connection):
        if not connection.upload.is_choked():
            self._rechoke()

    def not_interested(self, connection):
        if not connection.upload.is_choked():
            self._rechoke()

    def set_super_seed(self):
        while self.connections:             # close all connections
            self.connections[0].close()
        self.picker.set_superseed()
        self.super_seed = True

    def pause(self, flag):
        self.paused = flag
        self._rechoke()
