# Written by John Hoffman
# see LICENSE.txt for license information
# Updated and modified for ABC_OKC : Old King Cole

from time import clock
from math import log


class SingleBuffer:
    def __init__(self, pool, size):
        self.bufferpool = pool
        self.size = size
        self.buf = bytearray(size)
        self.bufmv = memoryview(self.buf)
        self.lastrelease = 0

    def init(self):
        self.length = 0

    def append(self, s):
        l = self.length + len(s)
        self.buf[self.length:l] = s
        self.length = l

    def __len__(self):
        return self.length

    def __getslice__(self, a, b):
        if b > self.length:
            b = self.length
        elif b < 0:
            b += self.length
        return self.bufmv[a:b]

    def getArrayCopy(self, slice = None):
        if slice is None:
            begin = 0
            end = self.length
        else:
            begin, end = slice
        return str(self.buf[begin:end])

    def release(self):
        self.lastrelease = clock()
        self.bufferpool.release(self)


class BufferPool:
    def __init__(self, sched, config):
        self.piecepool = {}
        self.sched = sched
        self.read_buffer_pool_expiration = config['read_buffer_pool_expiration']
        self.sched(self.expireBuffer, 10)

    def new(self, size):
        if size <= 32768:
            bufsize= 32768
        else:
            bufsize = 2 ** (int(log(size, 2)) + 1)
        if bufsize not in self.piecepool:
            self.piecepool[bufsize] = []
        if self.piecepool[bufsize]:
            x = self.piecepool[bufsize].pop(0)
        else:
            x = SingleBuffer(self, bufsize)
        x.init()
        return x

    def release(self, x):
        self.piecepool[x.size].append(x)

    def clear(self):
        for bufsize in self.piecepool:
            del self.piecepool[bufsize][:]

    def expireBuffer(self):
        now = clock()
        for bufsize in self.piecepool:
            for buffer in self.piecepool[bufsize][:]:
                if now - buffer.lastrelease > self.read_buffer_pool_expiration:
                    self.piecepool[bufsize].remove(buffer)
        self.sched(self.expireBuffer, 10)
